# Create Manual using doctools
#
# (cmd) gendoc format infile
# format : html latex , normal
# (Ex)  tclsh gendoc.tcl html test.man > test.html
package require doctools
doctools::new doc

set fmt    [lindex $argv 0]
set infile [lindex $argv 1]


if {[catch {
	if {$fmt eq "html"} {
	set meta {
<meta HTTP-EQUIV="content-type" CONTENT="text/html; charset=UTF-8">
<style type="text/css">
h2 {
	font-family: "Arial";
	font-weight: bolder;
	background-color: #CCCC00;
}
h1 {
	font-family: "Arial";
	font-weight: 800;
}
.cmd {
	font-style: bold;
}
.arg {
    color: #5555FF;
    font-style: italic;
}
</style>}
	} else {
		set meta ""
	}

	fconfigure stdout -encoding utf-8
	doc configure -format $fmt
	doc configure -file [file tail $infile]
	doc setparam meta $meta
#	doc setparam xref [list {kw,LaTeX fname fragment} ]
}]} {
	puts {error noformat}
	exit 1
}


if {[catch {open $infile r} in]} {
	puts {error: no file}
	exit 1
}
fconfigure $in -encoding utf-8


if {[catch {puts [doc format [read $in]]} msg]} {
	close $in
	puts $msg
	exit 1 
}


close $in
exit 0
