#include <windows.h>
#include <tcl.h>


typedef HRESULT (APIENTRY *LPCONVERTINETSTRING)(LPDWORD, DWORD, DWORD, LPCSTR, LPINT, LPBYTE, LPINT);
typedef HRESULT (APIENTRY *LPISCONVERTINETSTRINGAVALABLE)(DWORD, DWORD);

LPCONVERTINETSTRING ConvertINetString = NULL;
LPISCONVERTINETSTRINGAVALABLE IsConvertINetStringAvailable = NULL;
static HMODULE hDLL = NULL;


static BOOL InitDLL();
static int Mlang_EnumMimeList (Tcl_Interp *interp, char *regparent);
static int Mlang_GetAliasMap (Tcl_Interp *interp);
static int Mlang_GetCodepageFromAlias (Tcl_Interp *interp, Tcl_Obj *obj, DWORD *cp);
static int Mlang_GetCodePage (Tcl_Interp *interp, Tcl_Obj *obj, DWORD *cp);
static int Mlang_CheckConvertError(Tcl_Interp *interp, HRESULT result);
static int Mlang_EncodingObjCmd ( ClientData data, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]);

static BOOL
InitDLL()
{
    if(hDLL == NULL){
        hDLL = LoadLibrary("mlang.dll");
    }
    if(hDLL == NULL) return FALSE;

    ConvertINetString = (LPCONVERTINETSTRING)
        GetProcAddress(hDLL, "ConvertINetString");
    IsConvertINetStringAvailable = (LPISCONVERTINETSTRINGAVALABLE)
        GetProcAddress(hDLL, "IsConvertINetStringAvailable");

    if (!(ConvertINetString && IsConvertINetStringAvailable)) 
        return FALSE;

	return TRUE;
}


static int
Mlang_EnumMimeList (Tcl_Interp *interp, char *regparent)
{
    LONG result;
    HKEY hKey;
    DWORD index;
    char name[256];
    DWORD  count;
    FILETIME writetime;
    Tcl_Obj *tclList;
	Tcl_DString ds;

    Tcl_DStringInit(&ds);
    Tcl_DStringAppend(&ds, "MIME\\Database\\", -1);
    Tcl_DStringAppend(&ds, regparent, -1);

	if( RegOpenKeyEx(HKEY_CLASSES_ROOT, Tcl_DStringValue(&ds), 
                0, KEY_READ, &hKey) != ERROR_SUCCESS ){
        Tcl_DStringFree(&ds);
        Tcl_AppendResult(interp, "failed to open registry", NULL);
        return TCL_ERROR;
    }
    Tcl_DStringFree(&ds);
    
    tclList = Tcl_NewObj();
    index = 0;
    while (1) {
        count = sizeof(name);
        result = RegEnumKeyEx(hKey, index, name, &count, 0, 0, 0, &writetime);
        if (result != ERROR_SUCCESS) break;
        index++;

        Tcl_ListObjAppendElement(interp, tclList, 
                Tcl_NewStringObj(name, count));
    }

    RegCloseKey(hKey);

    Tcl_SetObjResult(interp, tclList);
    return TCL_OK;
}


static int
Mlang_GetAliasMap (Tcl_Interp *interp)
{
    int i, len;
    Tcl_Obj *charsetList;
    Tcl_Obj *resultList;
    DWORD cp;

    Mlang_EnumMimeList (interp, "Charset");
    charsetList = Tcl_GetObjResult(interp);

    resultList = Tcl_NewObj();
    Tcl_ListObjLength(interp, charsetList, &len);
    for (i=0; i < len; i++) {
        Tcl_Obj *obj;
        Tcl_ListObjIndex(interp, charsetList, i, &obj);
        if (Mlang_GetCodepageFromAlias(interp, obj, &cp) == TCL_OK) {
            Tcl_ListObjAppendElement(interp, resultList, obj);
            Tcl_ListObjAppendElement(interp, resultList, Tcl_NewIntObj(cp));
        }
    }

    Tcl_ResetResult(interp);
    Tcl_SetObjResult(interp, resultList);
    return TCL_OK;
}


static int
Mlang_GetCodepageFromAlias (Tcl_Interp *interp, Tcl_Obj *obj, DWORD *cp)
{
    LONG result;
    HKEY hKey = NULL;
	Tcl_DString ds;
    DWORD type, data, size;
    static char regcharset[] = "MIME\\Database\\Charset\\";

    Tcl_DStringInit(&ds);
    Tcl_DStringAppend(&ds, regcharset, -1);
    Tcl_DStringAppend(&ds, Tcl_GetString(obj), -1);

	result = RegOpenKeyEx(HKEY_CLASSES_ROOT, Tcl_DStringValue(&ds), 
            0, KEY_READ, &hKey);
	if(result != ERROR_SUCCESS){
        goto error;
    }
    
    size = sizeof(data);
    result = RegQueryValueEx(hKey, "InternetEncoding", 0, 
            &type, (LPBYTE)&data, &size);
	if(result != ERROR_SUCCESS || type != REG_DWORD){
        char strdata[256];
        size = sizeof(strdata);
        result = RegQueryValueEx(hKey, "AliasForCharset", 0, 
                &type, (LPBYTE)&strdata, &size);
        if (result != ERROR_SUCCESS || type != REG_SZ) {
            goto error;
        }
        if (Mlang_GetCodepageFromAlias(interp, 
                    Tcl_NewStringObj(strdata,-1), &data) != TCL_OK) {
            goto error;
        }
    }

    *cp = data;

    Tcl_DStringFree(&ds);
    RegCloseKey(hKey);
    return TCL_OK;
error:
    Tcl_DStringFree(&ds);
    if (hKey) RegCloseKey(hKey);
    Tcl_AppendResult(interp, "invalid alias name \"", 
            Tcl_GetString(obj), "\"", NULL);
    return TCL_ERROR;
}


static int
Mlang_GetCodePage (Tcl_Interp *interp, Tcl_Obj *obj, DWORD *cp)
{
    DWORD identifier;

    if (Mlang_GetCodepageFromAlias(interp, obj, &identifier) != TCL_OK) {
        if (Tcl_GetLongFromObj(interp, obj, &identifier) != TCL_OK) {
            Tcl_ResetResult(interp);
            Tcl_AppendResult(interp, "invalid codepage or alias name \"", 
                    Tcl_GetString(obj), "\"", NULL);
            return TCL_ERROR;
        }
    }
    *cp = identifier;

    Tcl_ResetResult(interp);
    return TCL_OK;
}

static int
Mlang_CheckConvertError(Tcl_Interp *interp, HRESULT result)
{
    switch (result) {
        case S_OK:
            return TCL_OK;
        case S_FALSE:
            Tcl_AppendResult(interp, "The specified conversion is not supported on the system", NULL);
            break;
            Tcl_AppendResult(interp, "Faied to convert", NULL);
        case E_FAIL:
            break;
    }
    return TCL_ERROR;
}


static int
Mlang_EncodingObjCmd (
    ClientData data,
    Tcl_Interp *interp,
    int objc,
    Tcl_Obj *CONST objv[])
{
    HRESULT result;
    DWORD mode = 0, from, to;
    int fromlen, tolen;
    char *frombyte = NULL, *tobyte = NULL;
    int index;

    static CONST char *optionStrings[] = {
    "aliasnames", "aliasmaps", "convert", "convertfrom", "convertto", 
    "names", NULL
    };
    enum options {
    MLANG_ENC_ALIASNAMES, MLANG_ENC_ALIASMAPS,
    MLANG_ENC_CONVERT, MLANG_ENC_CONVERTFROM, MLANG_ENC_CONVERTTO, 
    MLANG_ENC_NAMES, 
    };
    
    if (objc < 2) {
        Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
        return TCL_ERROR;
    }
    if (Tcl_GetIndexFromObj(interp, objv[1], optionStrings, "option", 0, &index) != TCL_OK) {
        return TCL_ERROR;
    }

    switch (index) {
        case MLANG_ENC_CONVERT:
            if (objc != 5) {
                Tcl_WrongNumArgs(interp, 2, objv, "fromEncoding toEncoding data");
                return TCL_ERROR;
            }

            if (Mlang_GetCodePage(interp, objv[2], &from) != TCL_OK) {
                return TCL_ERROR;
            }
            if (Mlang_GetCodePage(interp, objv[3], &to) != TCL_OK) {
                return TCL_ERROR;
            }

            frombyte = Tcl_GetByteArrayFromObj(objv[4], &fromlen);
            tolen = fromlen * 4;
            tobyte = Tcl_Alloc (tolen * sizeof(char));

            result = ConvertINetString (
                &mode, from, to,
                frombyte, &fromlen,
                tobyte, &tolen);
            
            if (result == S_OK) {
                Tcl_SetObjResult(interp, Tcl_NewByteArrayObj(tobyte, tolen));
            }
            if(tobyte) Tcl_Free(tobyte); tobyte = NULL; frombyte = NULL;

            if (Mlang_CheckConvertError(interp, result) != TCL_OK) {
                return TCL_ERROR;
            }

            break;

        case MLANG_ENC_CONVERTFROM:
        case MLANG_ENC_CONVERTTO:
            if (objc != 4) {
                Tcl_WrongNumArgs(interp, 2, objv, "encoding data");
                return TCL_ERROR;
            }
            if (index == MLANG_ENC_CONVERTFROM) {
                if (Mlang_GetCodePage(interp, objv[2], &from) != TCL_OK) {
                    return TCL_ERROR;
                }
                to = 65001;

                frombyte = Tcl_GetByteArrayFromObj(objv[3], &fromlen);
                tolen = fromlen * 4;
                tobyte = Tcl_Alloc (tolen * sizeof(char));

                result = ConvertINetString (
                    &mode, from, to,
                    frombyte, &fromlen,
                    tobyte, &tolen);
                if (result == S_OK) {
                    Tcl_SetObjResult(interp, Tcl_NewStringObj(tobyte, tolen));
                }
                if(tobyte) Tcl_Free(tobyte); tobyte = NULL; frombyte = NULL;

            } else {
                if (Mlang_GetCodePage(interp, objv[2], &to) != TCL_OK) {
                    return TCL_ERROR;
                }
                from = 65001;

                frombyte = Tcl_GetStringFromObj(objv[3], &fromlen);
                tolen = fromlen * 4;
                tobyte = Tcl_Alloc(tolen * sizeof(char));

                result = ConvertINetString (
                    &mode, from, to,
                    frombyte, &fromlen,
                    tobyte, &tolen);
                if (result == S_OK) {
                    Tcl_SetObjResult(interp, Tcl_NewByteArrayObj(tobyte, tolen));
                }
                if(tobyte) Tcl_Free(tobyte); tobyte = NULL; frombyte = NULL;
            }

            if (Mlang_CheckConvertError(interp, result) != TCL_OK) {
                return TCL_ERROR;
            }
            break;

        case MLANG_ENC_NAMES:
            if (objc > 2) {
                Tcl_WrongNumArgs(interp, 2, objv, NULL);
                return TCL_ERROR;
            }
            return Mlang_EnumMimeList (interp, "Codepage");

        case MLANG_ENC_ALIASNAMES:
            if (objc > 2) {
                Tcl_WrongNumArgs(interp, 2, objv, NULL);
                return TCL_ERROR;
            }
            return Mlang_EnumMimeList (interp, "Charset");

        case MLANG_ENC_ALIASMAPS:
            if (objc > 2) {
                Tcl_WrongNumArgs(interp, 2, objv, NULL);
                return TCL_ERROR;
            }
            return Mlang_GetAliasMap (interp);
    }
    return TCL_OK;
}


DLLEXPORT int
Mlenc_Init (Tcl_Interp *interp)
{
#ifdef USE_TCL_STUBS
	if (Tcl_InitStubs(interp, "8", 0) == NULL) {
		return TCL_ERROR;
	}
#endif
    if (InitDLL() == FALSE) {
        Tcl_AppendResult(interp, "failed to load the library", NULL);
        return TCL_ERROR;
    }
    if (Tcl_Eval(interp, "namespace eval ::" PACKAGE_NAME " {}") != TCL_OK) {
        return TCL_ERROR;
    }
	Tcl_CreateObjCommand(interp, "::" PACKAGE_NAME "::encoding", Mlang_EncodingObjCmd, NULL, NULL);

	if (Tcl_PkgProvide(interp, PACKAGE_NAME, PACKAGE_VERSION) != TCL_OK) {
		return TCL_ERROR;
	}
	return TCL_OK;
}


DLLEXPORT int
Mlenc_Unload (Tcl_Interp *interp)
{
    if (Tcl_Eval(interp, "namespace delete ::" PACKAGE_NAME) != TCL_OK) {
        return TCL_ERROR;
    }
    FreeLibrary(hDLL);

    hDLL = NULL;
    ConvertINetString = NULL;
    IsConvertINetStringAvailable = NULL;
    return TCL_OK;
}
